@@ -13,7 +13,6 @@
 use std::{
     cmp,
     collections::HashMap,
-    convert::identity,
     fmt, mem,
     ops::ControlFlow,
     panic::{self, AssertUnwindSafe},
@@ -308,11 +307,7 @@ impl<World, Which, Before, After> Basic<World, Which, Before, After> {
     /// [`Step`]s, even after [`Skipped`] of [`Failed`] ones.
     ///
     /// Last `World` argument is supplied to the function, in case it was
-    /// initialized before by running [`before`] hook or any non-failed
-    /// [`Step`]. In case the last [`Scenario`]'s [`Step`] failed, we want to
-    /// return event with an exact `World` state. Also, we don't want to impose
-    /// additional [`Clone`] bounds on `World`, so the only option left is to
-    /// pass [`None`] to the function.
+    /// initialized before by running [`before`] hook or any [`Step`].
     ///
     /// [`before`]: Self::before()
     /// [`Failed`]: event::Step::Failed
@@ -735,28 +730,19 @@ where
                 event::Cucumber::scenario(f, r, s, e(step, captures))
             }
         };
-        let err = |e: fn(_, _, _, _) -> event::Scenario<W>| {
-            let (f, r, s) = (&feature, &rule, &scenario);
-            move |step, captures, w, info| {
-                let (f, r, s) = (Arc::clone(f), r.clone(), Arc::clone(s));
-                event::Cucumber::scenario(f, r, s, e(step, captures, w, info))
-            }
-        };
 
-        let compose = |started, passed, skipped, failed| {
-            (ok(started), ok_capt(passed), ok(skipped), err(failed))
+        let compose = |started, passed, skipped| {
+            (ok(started), ok_capt(passed), ok(skipped))
         };
         let into_bg_step_ev = compose(
             event::Scenario::background_step_started,
             event::Scenario::background_step_passed,
             event::Scenario::background_step_skipped,
-            event::Scenario::background_step_failed,
         );
         let into_step_ev = compose(
             event::Scenario::step_started,
             event::Scenario::step_passed,
             event::Scenario::step_skipped,
-            event::Scenario::step_failed,
         );
 
         self.send_event(event::Cucumber::scenario(
@@ -766,12 +752,10 @@ where
             event::Scenario::Started,
         ));
 
-        let mut is_failed = false;
-        let world = async {
+        let mut result = async {
             let before_hook = self
                 .run_before_hook(&feature, rule.as_ref(), &scenario)
-                .await
-                .map_err(|_unit| None)?;
+                .await?;
 
             let feature_background = feature
                 .background
@@ -783,7 +767,8 @@ where
             let feature_background = stream::iter(feature_background)
                 .map(Ok)
                 .try_fold(before_hook, |world, bg_step| {
-                    self.run_step(world, bg_step, into_bg_step_ev).map_ok(Some)
+                    self.run_step(world, bg_step, true, into_bg_step_ev)
+                        .map_ok(Some)
                 })
                 .await?;
 
@@ -802,28 +787,53 @@ where
             let rule_background = stream::iter(rule_background)
                 .map(Ok)
                 .try_fold(feature_background, |world, bg_step| {
-                    self.run_step(world, bg_step, into_bg_step_ev).map_ok(Some)
+                    self.run_step(world, bg_step, true, into_bg_step_ev)
+                        .map_ok(Some)
                 })
                 .await?;
 
             stream::iter(scenario.steps.iter().map(|s| Arc::new(s.clone())))
                 .map(Ok)
                 .try_fold(rule_background, |world, step| {
-                    self.run_step(world, step, into_step_ev).map_ok(Some)
+                    self.run_step(world, step, false, into_step_ev).map_ok(Some)
                 })
                 .await
         }
-        .inspect_err(|e| {
-            if e.is_none() {
-                is_failed = true;
-            }
-        })
-        .await
-        .unwrap_or_else(identity);
+        .await;
 
-        self.run_after_hook(world, &feature, rule.as_ref(), &scenario)
+        let world = match &mut result {
+            Ok(world) => world.take(),
+            Err(exec_err) => exec_err.take_world(),
+        };
+
+        let (world, after_hook_meta, after_hook_error) = self
+            .run_after_hook(world, &feature, rule.as_ref(), &scenario)
             .await
-            .map_or_else(|_| is_failed = true, drop);
+            .map_or_else(
+                |(w, meta, info)| (w.map(Arc::new), Some(meta), Some(info)),
+                |(w, meta)| (w.map(Arc::new), meta, None),
+            );
+
+        let is_failed = result.is_err() || after_hook_error.is_some();
+
+        if let Some(exec_error) = result.err() {
+            self.emit_failed_events(
+                Arc::clone(&feature),
+                rule.clone(),
+                Arc::clone(&scenario),
+                world.clone(),
+                exec_error,
+            );
+        }
+
+        self.emit_after_hook_events(
+            Arc::clone(&feature),
+            rule.clone(),
+            Arc::clone(&scenario),
+            world,
+            after_hook_meta,
+            after_hook_error,
+        );
 
         self.send_event(event::Cucumber::scenario(
             Arc::clone(&feature),
@@ -839,13 +849,16 @@ where
     ///
     /// # Events
     ///
-    /// - Emits [`HookType::Before`] event.
+    /// - Emits all the [`HookType::Before`] events, except [`Hook::Failed`].
+    ///   See [`Self::emit_failed_events()`] for more details.
+    ///
+    /// [`Hook::Failed`]: event::Hook::Failed
     async fn run_before_hook(
         &self,
         feature: &Arc<gherkin::Feature>,
         rule: Option<&Arc<gherkin::Rule>>,
         scenario: &Arc<gherkin::Scenario>,
-    ) -> Result<Option<W>, ()> {
+    ) -> Result<Option<W>, ExecutionFailure<W>> {
         let init_world = async {
             AssertUnwindSafe(W::new())
                 .catch_unwind()
@@ -895,81 +908,12 @@ where
                     ));
                     Ok(Some(world))
                 }
-                Err((info, world)) => {
-                    self.send_event(event::Cucumber::scenario(
-                        Arc::clone(feature),
-                        rule.map(Arc::clone),
-                        Arc::clone(scenario),
-                        event::Scenario::hook_failed(
-                            HookType::Before,
-                            world.map(Arc::new),
-                            info,
-                        ),
-                    ));
-                    Err(())
-                }
-            }
-        } else {
-            Ok(None)
-        }
-    }
-
-    /// Executes [`HookType::After`], if present.
-    ///
-    /// # Event
-    ///
-    /// - Emits [`HookType::After`] event.
-    async fn run_after_hook(
-        &self,
-        mut world: Option<W>,
-        feature: &Arc<gherkin::Feature>,
-        rule: Option<&Arc<gherkin::Rule>>,
-        scenario: &Arc<gherkin::Scenario>,
-    ) -> Result<Option<W>, ()> {
-        if let Some(hook) = self.after_hook.as_ref() {
-            self.send_event(event::Cucumber::scenario(
-                Arc::clone(feature),
-                rule.map(Arc::clone),
-                Arc::clone(scenario),
-                event::Scenario::hook_started(HookType::After),
-            ));
-
-            let fut = async {
-                let fut = (hook)(
-                    feature.as_ref(),
-                    rule.as_ref().map(AsRef::as_ref),
-                    scenario.as_ref(),
-                    world.as_mut(),
-                );
-                match AssertUnwindSafe(fut).catch_unwind().await {
-                    Ok(()) => Ok(world),
-                    Err(info) => Err((info, world)),
-                }
-            };
-
-            #[allow(clippy::shadow_unrelated)]
-            match fut.await {
-                Ok(world) => {
-                    self.send_event(event::Cucumber::scenario(
-                        Arc::clone(feature),
-                        rule.map(Arc::clone),
-                        Arc::clone(scenario),
-                        event::Scenario::hook_passed(HookType::After),
-                    ));
-                    Ok(world)
-                }
-                Err((info, world)) => {
-                    self.send_event(event::Cucumber::scenario(
-                        Arc::clone(feature),
-                        rule.map(Arc::clone),
-                        Arc::clone(scenario),
-                        event::Scenario::hook_failed(
-                            HookType::After,
-                            world.map(Arc::new),
-                            info.into(),
-                        ),
-                    ));
-                    Err(())
+                Err((panic_info, world)) => {
+                    Err(ExecutionFailure::BeforeHookPanicked {
+                        world,
+                        panic_info,
+                        meta: event::Metadata::new(()),
+                    })
                 }
             }
         } else {
@@ -981,25 +925,22 @@ where
     ///
     /// # Events
     ///
-    /// - Emits all [`Step`] events.
+    /// - Emits all the [`Step`] events, except [`Step::Failed`]. See
+    ///   [`Self::emit_failed_events()`] for more details.
     ///
     /// [`Step`]: gherkin::Step
-    async fn run_step<St, Ps, Sk, F>(
+    /// [`Step::Failed`]: event::Step::Failed
+    async fn run_step<St, Ps, Sk>(
         &self,
         world: Option<W>,
         step: Arc<gherkin::Step>,
-        (started, passed, skipped, failed): (St, Ps, Sk, F),
-    ) -> Result<W, Option<W>>
+        is_background: bool,
+        (started, passed, skipped): (St, Ps, Sk),
+    ) -> Result<W, ExecutionFailure<W>>
     where
         St: FnOnce(Arc<gherkin::Step>) -> event::Cucumber<W>,
         Ps: FnOnce(Arc<gherkin::Step>, CaptureLocations) -> event::Cucumber<W>,
         Sk: FnOnce(Arc<gherkin::Step>) -> event::Cucumber<W>,
-        F: FnOnce(
-            Arc<gherkin::Step>,
-            Option<CaptureLocations>,
-            Option<Arc<W>>,
-            event::StepError,
-        ) -> event::Cucumber<W>,
     {
         self.send_event(started(Arc::clone(&step)));
 
@@ -1053,20 +994,185 @@ where
             }
             Ok((_, world)) => {
                 self.send_event(skipped(step));
-                Err(world)
+                Err(ExecutionFailure::StepSkipped(world))
             }
             Err((err, captures, world)) => {
-                self.send_event(failed(
+                Err(ExecutionFailure::StepPanicked {
+                    world,
                     step,
                     captures,
-                    world.map(Arc::new),
                     err,
-                ));
-                Err(None)
+                    meta: event::Metadata::new(()),
+                    is_background,
+                })
             }
         }
     }
 
+    /// Emits all the failure events of [`HookType::Before`] or [`Step`] after
+    /// executing the [`Self::run_after_hook()`].
+    ///
+    /// This is done because [`HookType::After`] requires a mutable reference to
+    /// the [`World`] while on the other hand we store immutable reference to it
+    /// inside failure events for easier debugging. So, to avoid imposing
+    /// additional [`Clone`] bounds on the [`World`], we run the
+    /// [`HookType::After`] first without emitting any events about its
+    /// execution, then emit failure event of the [`HookType::Before`] or
+    /// [`Step`], if present, and finally emit all the [`HookType::After`]
+    /// events. This allows us to ensure [order guarantees][1] while not
+    /// restricting the [`HookType::After`] to the immutable reference. The only
+    /// downside of this approach is that we may emit failure events of
+    /// [`HookType::Before`] or [`Step`] with the [`World`] state being changed
+    /// by the [`HookType::After`].
+    ///
+    /// [`Step`]: gherkin::Step
+    /// [1]: crate::Runner#order-guarantees
+    fn emit_failed_events(
+        &self,
+        feature: Arc<gherkin::Feature>,
+        rule: Option<Arc<gherkin::Rule>>,
+        scenario: Arc<gherkin::Scenario>,
+        world: Option<Arc<W>>,
+        err: ExecutionFailure<W>,
+    ) {
+        match err {
+            ExecutionFailure::StepSkipped(_) => {}
+            ExecutionFailure::BeforeHookPanicked {
+                panic_info, meta, ..
+            } => {
+                self.send_event_with_meta(
+                    event::Cucumber::scenario(
+                        feature,
+                        rule,
+                        scenario,
+                        event::Scenario::hook_failed(
+                            HookType::Before,
+                            world,
+                            panic_info,
+                        ),
+                    ),
+                    meta,
+                );
+            }
+            ExecutionFailure::StepPanicked {
+                step,
+                captures,
+                err: error,
+                meta,
+                is_background: true,
+                ..
+            } => self.send_event_with_meta(
+                event::Cucumber::scenario(
+                    feature,
+                    rule,
+                    scenario,
+                    event::Scenario::background_step_failed(
+                        step, captures, world, error,
+                    ),
+                ),
+                meta,
+            ),
+            ExecutionFailure::StepPanicked {
+                step,
+                captures,
+                err: error,
+                meta,
+                is_background: false,
+                ..
+            } => self.send_event_with_meta(
+                event::Cucumber::scenario(
+                    feature,
+                    rule,
+                    scenario,
+                    event::Scenario::step_failed(step, captures, world, error),
+                ),
+                meta,
+            ),
+        }
+    }
+
+    /// Executes the [`HookType::After`], if present.
+    ///
+    /// Doesn't emit any events, see [`Self::emit_failed_events()`] for more
+    /// details.
+    async fn run_after_hook(
+        &self,
+        mut world: Option<W>,
+        feature: &Arc<gherkin::Feature>,
+        rule: Option<&Arc<gherkin::Rule>>,
+        scenario: &Arc<gherkin::Scenario>,
+    ) -> Result<
+        (Option<W>, Option<AfterHookEventsMeta>),
+        (Option<W>, AfterHookEventsMeta, Info),
+    > {
+        if let Some(hook) = self.after_hook.as_ref() {
+            let fut = (hook)(
+                feature.as_ref(),
+                rule.as_ref().map(AsRef::as_ref),
+                scenario.as_ref(),
+                world.as_mut(),
+            );
+
+            let started = event::Metadata::new(());
+            let res = AssertUnwindSafe(fut).catch_unwind().await;
+            let finished = event::Metadata::new(());
+            let meta = AfterHookEventsMeta { started, finished };
+
+            match res {
+                Ok(()) => Ok((world, Some(meta))),
+                Err(info) => Err((world, meta, info.into())),
+            }
+        } else {
+            Ok((world, None))
+        }
+    }
+
+    /// Emits all the [`HookType::After`] events.
+    ///
+    /// See [`Self::emit_failed_events()`] for the explanation why we don't do
+    /// that inside [`Self::run_after_hook()`].
+    fn emit_after_hook_events(
+        &self,
+        feature: Arc<gherkin::Feature>,
+        rule: Option<Arc<gherkin::Rule>>,
+        scenario: Arc<gherkin::Scenario>,
+        world: Option<Arc<W>>,
+        meta: Option<AfterHookEventsMeta>,
+        err: Option<Info>,
+    ) {
+        debug_assert_eq!(self.after_hook.is_some(), meta.is_some());
+
+        if let Some(meta) = meta {
+            self.send_event_with_meta(
+                event::Cucumber::scenario(
+                    Arc::clone(&feature),
+                    rule.clone(),
+                    Arc::clone(&scenario),
+                    event::Scenario::hook_started(HookType::After),
+                ),
+                meta.started,
+            );
+
+            let ev = if let Some(err) = err {
+                event::Cucumber::scenario(
+                    feature,
+                    rule,
+                    scenario,
+                    event::Scenario::hook_failed(HookType::After, world, err),
+                )
+            } else {
+                event::Cucumber::scenario(
+                    feature,
+                    rule,
+                    scenario,
+                    event::Scenario::hook_passed(HookType::After),
+                )
+            };
+
+            self.send_event_with_meta(ev, meta.finished);
+        }
+    }
+
     /// Notifies [`FinishedRulesAndFeatures`] about [`Scenario`] being finished.
     ///
     /// [`Scenario`]: gherkin::Scenario
@@ -1088,11 +1194,25 @@ where
     ///
     /// [`Cucumber`]: event::Cucumber
     fn send_event(&self, event: event::Cucumber<W>) {
-        // If the receiver end is dropped, then no one listens for events
+        // If the receiver end is dropped, then no one listens for events,
         // so we can just ignore it.
         drop(self.event_sender.unbounded_send(Ok(Event::new(event))));
     }
 
+    /// Notifies with the given [`Cucumber`] event along with its [`Metadata`].
+    ///
+    /// [`Cucumber`]: event::Cucumber
+    /// [`Metadata`]: event::Metadata
+    fn send_event_with_meta(
+        &self,
+        event: event::Cucumber<W>,
+        meta: event::Metadata,
+    ) {
+        // If the receiver end is dropped, then no one listens for events,
+        // so we can just ignore it.
+        drop(self.event_sender.unbounded_send(Ok(meta.wrap(event))));
+    }
+
     /// Notifies with the given [`Cucumber`] events.
     ///
     /// [`Cucumber`]: event::Cucumber
@@ -1101,7 +1221,7 @@ where
         events: impl Iterator<Item = event::Cucumber<W>>,
     ) {
         for v in events {
-            // If the receiver end is dropped, then no one listens for events
+            // If the receiver end is dropped, then no one listens for events,
             // so we can just stop from here.
             if self.event_sender.unbounded_send(Ok(Event::new(v))).is_err() {
                 break;
@@ -1417,3 +1537,93 @@ impl Features {
 fn coerce_into_info<T: std::any::Any + Send + 'static>(val: T) -> Info {
     Arc::new(val)
 }
+
+/// Failure encountered during execution of [`HookType::Before`] or [`Step`].
+/// See [`Executor::emit_failed_events()`] for more info.
+///
+/// [`Step`]: gherkin::Step
+enum ExecutionFailure<World> {
+    /// [`HookType::Before`] panicked.
+    BeforeHookPanicked {
+        /// [`World`] at the time [`HookType::Before`] has panicked.
+        world: Option<World>,
+
+        /// [`catch_unwind()`] of the [`HookType::Before`] panic.
+        ///
+        /// [`catch_unwind()`]: std::panic::catch_unwind
+        panic_info: Info,
+
+        /// [`Metadata`] at the time [`HookType::Before`] panicked.
+        ///
+        /// [`Metadata`]: event::Metadata
+        meta: event::Metadata,
+    },
+
+    /// [`Step`] was skipped.
+    ///
+    /// [`Step`]: gherkin::Step.
+    StepSkipped(Option<World>),
+
+    /// [`Step`] failed.
+    ///
+    /// [`Step`]: gherkin::Step.
+    StepPanicked {
+        /// [`World`] at the time when [`Step`] has failed.
+        ///
+        /// [`Step`]: gherkin::Step
+        world: Option<World>,
+
+        /// [`Step`] itself.
+        ///
+        /// [`Step`]: gherkin::Step
+        step: Arc<gherkin::Step>,
+
+        /// [`Step`]s [`regex`] [`CaptureLocations`].
+        ///
+        /// [`Step`]: gherkin::Step
+        captures: Option<CaptureLocations>,
+
+        /// [`StepError`] of the [`Step`].
+        ///
+        /// [`Step`]: gherkin::Step
+        /// [`StepError`]: event::StepError
+        err: event::StepError,
+
+        /// [`Metadata`] at the time when [`Step`] failed.
+        ///
+        /// [`Metadata`]: event::Metadata
+        /// [`Step`]: gherkin::Step.
+        meta: event::Metadata,
+
+        /// Indicator whether the [`Step`] was background or not.
+        ///
+        /// [`Step`]: gherkin::Step
+        is_background: bool,
+    },
+}
+
+/// [`Metadata`] of [`HookType::After`] events.
+///
+/// [`Metadata`]: event::Metadata
+struct AfterHookEventsMeta {
+    /// [`Metadata`] at the time [`HookType::After`] started.
+    ///
+    /// [`Metadata`]: event::Metadata
+    started: event::Metadata,
+
+    /// [`Metadata`] at the time [`HookType::After`] finished.
+    ///
+    /// [`Metadata`]: event::Metadata
+    finished: event::Metadata,
+}
+
+impl<W> ExecutionFailure<W> {
+    /// Takes the [`World`] leaving a [`None`] in its place.
+    fn take_world(&mut self) -> Option<W> {
+        match self {
+            Self::BeforeHookPanicked { world, .. }
+            | Self::StepSkipped(world)
+            | Self::StepPanicked { world, .. } => world.take(),
+        }
+    }
+}
